#!/usr/bin/env python
# -*- coding: utf-8 -*-

import glob
import json
import logging
import os

import torch
from explaining_framework.utils.io import read_yaml
from torch_geometric.graphgym.model_builder import create_model
from torch_geometric.graphgym.train import GraphGymDataModule
from torch_geometric.graphgym.utils.io import json_to_dict_list

MODEL_STATE = "model_state"
OPTIMIZER_STATE = "optimizer_state"
SCHEDULER_STATE = "scheduler_state"


def _load_ckpt(
    model: torch.nn.Module,
    ckpt_path: str,
) -> torch.nn.Module:
    r"""Loads the model at given checkpoint."""

    if not os.path.exists(ckpt_path):
        return None

    ckpt = torch.load(ckpt_path)
    model.load_state_dict(ckpt["state_dict"])

    return model


class LoadModelInfo(object):
    def __init__(
        self, dataset_name: str, model_dir: str, which: str, wrt_metric: str = "val"
    ):
        self.dataset_name = dataset_name
        self.model_dir = model_dir
        self.which = which
        self.wrt_metric = wrt_metric
        self.info = None

    def list_stats(self, path) -> list:
        info = []
        for path_ in glob.glob(
            os.path.join(path, "[0-9]", self.wrt_metric, "stats.json")
        ):
            stats = json_to_dict_list(path_)
            for stat in stats:
                xp_dir_path = os.path.dirname(os.path.dirname(os.path.dirname(path_)))
                seed = int(os.path.basename(os.path.dirname(os.path.dirname(path_))))
                ckpt_dir_path = os.path.join(
                    os.path.dirname(os.path.dirname(path_)), "ckpt"
                )
                cfg_path = os.path.join(xp_dir_path, "config.yaml")
                epoch = stat["epoch"]
                accuracy = stat["accuracy"]
                loss = stat["loss"]
                lr = stat["lr"]
                params = stat["params"]
                time_iter = stat["time_iter"]
                ckpt_path = self.get_ckpt_path(epoch=epoch, ckpt_dir_path=ckpt_dir_path)
                info.append(
                    {
                        "xp_dir_path": xp_dir_path,
                        "ckpt_path": self.get_ckpt_path(
                            epoch=epoch, ckpt_dir_path=ckpt_dir_path
                        ),
                        "cfg_path": cfg_path,
                        "seed": seed,
                        "epoch": epoch,
                        "accuracy": accuracy,
                        "loss": loss,
                        "lr": lr,
                        "params": params,
                        "time_iter": time_iter,
                        "which": self.which
                        if self.which in ["best", "worst"]
                        else None,
                    }
                )
        return info

    def list_xp(self):
        paths = []
        for path in glob.glob(os.path.join(self.model_dir, "**", "config.yaml")):
            file = self.load_cfg(path)
            dataset_name_ = file["dataset"]["name"]
            if self.dataset_name == dataset_name_:
                paths.append(os.path.dirname(path))
        if len(paths) == 0:
            logging.warning(
                f"It does not exist any model trained for the dataset {self.dataset_name}"
            )
            return None
        else:
            return paths

    def load_cfg(self, config_path):
        return read_yaml(config_path)

    def set_info(self):
        if self.which in ["best", "worst"] or isinstance(self.which, int):
            paths = self.list_xp()
            infos = []
            for path in paths:
                info = self.list_stats(path)
                infos.extend(info)
            infos = sorted(infos, key=lambda item: item["accuracy"])
            if self.which == "best":
                self.info = infos[-1]
            elif self.which == "worst":
                self.info = infos[0]
            elif isinstance(self.which, int):
                self.info = infos[self.which]
        else:
            specific_xp_dir_path = os.path.dirname(
                os.path.dirname(os.path.dirname(self.which))
            )
            stats = self.list_stats(specific_xp_dir_path)
            self.info = [item for item in stats if item["ckpt_path"] == self.which][0]
        return self.info

    def get_model_signature(self):
        if self.info is None:
            self.set_info()

        model_name = os.path.basename(self.info["xp_dir_path"])
        model_seed = self.info["seed"]
        model_signature = "-".join(
            [
                f"{name}={val}"
                for name, val in zip(["name", "seed"], [model_name, model_seed])
            ]
            + [self.which]
        )
        return model_signature

    def get_ckpt_path(self, epoch: int, ckpt_dir_path: str):
        paths = os.path.join(ckpt_dir_path, "*.ckpt")
        ckpts = []
        for path in glob.glob(paths):
            feat = os.path.basename(path)
            feat = feat.split("-")
            for fe in feat:
                if "epoch" in fe:
                    fe_ep = int(fe.split("=")[1])
                    ckpts.append({"path": path, "div": abs(fe_ep - epoch)})
        ckpt = sorted(ckpts, key=lambda item: item["div"])[0]
        return ckpt["path"]
